module ani_steepds_illu_mod

  use sep
  use ani_type
  use ani_parms_cg
  use ani_exec_time
  use ani_taper
  use ani_wemvacg_illu
  use igrad1
  use base_spline_mod

  implicit none

  integer, private                                       :: iter
  complex, allocatable, private                          :: source1(:)
  real, allocatable, private                             :: taper(:)

  contains

  !======================================================================
  subroutine steepds_init(src_in, taper_in)

    complex                             :: src_in(:)
    real                                :: taper_in(:)

    if (allocated(source1)) deallocate(source1)
    if (allocated(taper))  deallocate(taper)

    allocate(source1(sou%w%n), taper(nx))

    source1 = src_in
    taper  = taper_in

  end subroutine steepds_init
  !======================================================================

  !======================================================================
  subroutine fgeval_init(iter_in)

    integer                             :: iter_in
    iter = iter_in

  end subroutine fgeval_init
  !======================================================================

  !======================================================================
  subroutine feval(model_in, refr_in, fun, Bimage, illumination)

    real                                                :: model_in(:,:,:), refr_in(:,:,:)
    real                                                :: Bimage(:,:,:)
    real, optional                                      :: illumination(:,:,:)
    double precision                                    :: fun

    real, allocatable                                   :: image(:,:,:), illum(:,:,:)
    complex, allocatable				:: recv(:,:),source(:,:)
    integer						:: ish, sh_position, iz, ix, isecdim, ipar, stat, stat1
    real                                                :: subh, start, finish

    allocate(source(rec%h%n, sou%w%n))
    allocate(recv(rec%h%n, rec%w%n))
    allocate(image(nx,nsecdim,nz))
    if (present(illumination)) then 
      allocate(illum(nx,nsecdim,nz))
    end if

    call sseek("rec",0,0)
    call aniwei_init()

    start=0; finish=0; 
    call aniexec_time(start)

    Bimage=0.; illumination=0. 
    do ish = 1, nsh
      sh_position = ((ish-1)*dsh + osh - rec%h%o) / dx + 1
      source=0.; source(sh_position,:) = conjg(source1);
      recv=0; call sreed("rec", recv, 8*rec%h%n*rec%w%n); recv = conjg(recv)
      image=0; 
      if (present(illumination)) then
        illum=0.; call aniwei(taper,recv,source,model_in,refr_in,image,illum)
        Bimage= Bimage + image
        illumination = illumination + illum 
      else
        call aniwei(taper,recv,source,model_in,refr_in,image)
        Bimage= Bimage + image
      end if
    end do

    if (present(illumination)) then
      where(illumination<illu_min)
        illumination = illu_min
      end where
      where(illumination>illu_max)
        illumination = illu_max
      end where
      ! Balance the Background image using illumination
      Bimage = Bimage / illumination 
    end if

    call aniexec_time(finish)
    write(0,112) "COMPUTED BACKGROUND IMAGE ", (finish-start)/60," min"         

    fun=dble(0.)
    do iz=1,nz
      do isecdim=1,nsecdim
        subh = ( (isecdim-1)*dsecdim+osecdim ) / ( (nsecdim-1)*dsecdim+osecdim )
        do ix=1,nx
          fun = fun + Bimage(ix,isecdim,iz)*Bimage(ix,isecdim,iz)*subh*subh
        end do
      end do
    end do

    write(0,*) "Objective fun = ", fun

    if(allocated(illum)) deallocate(illum)
    112 format(a29,f10.2,a20,i4)
  end subroutine feval

  !======================================================================
  !======================================================================

  subroutine geval(model_in, refr_in, Bimage, grad)

    real                                                :: model_in(:,:,:), refr_in(:,:,:)
    real                                                :: Bimage(:,:,:), grad(:)
    double precision                                    :: fun

    integer						:: ish, sh_position, iz, ix, isecdim, ipar, stat, stat1
    real, allocatable					:: mute(:), mute_init(:,:,:)
    double precision, allocatable                       :: temp_v(:), temp_e(:)
    real                                                :: start, finish
    real, allocatable                                   :: dmodel(:)
    double precision, allocatable                       :: gspl_v(:), gspl_e(:)

    ! Read in the mask for seabed 
    allocate(mute_init(nx,nz,npar), mute(nx*nz*npar))
    call sseek("mask",0,0)
    call sreed("mask", mute_init, 4*nx*nz*npar)
    mute = reshape(mute_init, (/nx*nz*npar/))
    deallocate(mute_init)

    ! Compute the gradient using the adjoint wemva
    start=0; finish=0; 
    call aniexec_time(start)
    call aniwemva_init(source1, taper, model_in, refr_in)
    allocate(dmodel(nz*nx*npar)); dmodel=0.

    ! compute the gradient
    call aniwemva_lop2(.true., .false., dmodel, Bimage)

    ! Scale and Smooth the gradient field
    allocate(temp_v(nx*nz), temp_e(nx*nz))
    allocate(gspl_v(axk_v%n*azk_v%n), gspl_e(axk_e%n*azk_e%n))
    call srite("grad_ns", dmodel, 4*nx*nz*npar) 
    ! Mute the dmodel down to seabed
    dmodel = dmodel * mute

    temp_v = dmodel(1:nx*nz);
    temp_e = dmodel(nx*nz+1:);
    gspl_v = 0.; gspl_e = 0.;

    ! Smooth velocity
    stat = base_spln2d_init(img%z, img%xm, azk_v, axk_v)
    stat1 = base_spln2d_oper(.true. , .false., gspl_v, temp_v)     
    stat1 = base_spln2d_oper(.false., .false., gspl_v, temp_v)

    ! Smooth eta 
    stat = base_spln2d_init(img%z, img%xm, azk_e, axk_e)
    stat1 = base_spln2d_oper(.true. , .false., gspl_e, temp_e)     
    stat1 = base_spln2d_oper(.false., .false., gspl_e, temp_e)

    ! Scale the gradient field
    temp_v = temp_v / maxval(abs(temp_v)) * 1e-4 * g_scales(1)
    temp_e = temp_e / maxval(abs(temp_e)) * 0.1 * g_scales(2)

    dmodel(1:nx*nz) = sngl(temp_v);
    dmodel(nx*nz+1:) = sngl(temp_e);

    grad = -dmodel

    call to_history("n4", iter, "grad")
    call srite("grad", dmodel, 4*nz*nx*npar)

    call aniexec_time(finish)
    write(0,112) "COMPUTED MODEL GRADIENT ",(finish-start)/60," min "
    if (allocated(dmodel)) deallocate(dmodel)
    if (allocated(temp_v)) deallocate(temp_v)
    if (allocated(temp_e)) deallocate(temp_e)
    if (allocated(gspl_v)) deallocate(gspl_v)
    if (allocated(gspl_e)) deallocate(gspl_e)

    deallocate( mute )

    111 format(a26,i4,a8,i4,a4,f6.2,a4)
    112 format(a29,f10.2,a20,i4)

  end subroutine geval
  !======================================================================

end module ani_steepds_illu_mod
